package TextData::File;

#textdata version 1.9.0 (19.8.2008)
#see POD below!

use TextData::Calc qw(LinearValue FormulaFunction);
use TextData::Lines;
use sort 'stable';

our @ISA = ('TextData::Lines');

sub new
{
	my $class = shift;
	my $config = shift;
	my $self = $class->SUPER::new($config);
	$self->{InFile} = shift;
	if(defined $self->{InFile})
	{
		$self->Read();
	}
	return $self;
}

sub Init
{
	#not touching InFile! this is only handled by Read
	#as is OutFile!
	my $self = shift;
	$self->SUPER::Init();
	$self->{config}{indexformat} = '%6E' unless defined $self->{config}->{indexformat};
	$self->{config}{extrapol} = 1 unless defined $self->{config}{extrapol};
	$self->{Data} = [];
	$self->{DataIndex} = [];
	$self->{SortedData} = [];
	$self->{RawHeader} = [];
	$self->{Buffer} = ""; #containing the raw lines read by readhead (the'd be lost from stdin otherwise)
}

sub ReadHead
{
	my $self = shift;
	$self->Init(); #do we really want full init?
	$self->{InFile} = shift if defined $_[0]; 
	if(defined $self->{InFile})
	{
		my $handle;
		if($self->{InFile} eq ''){ $handle = STDIN; }
		else{ open($handle, $self->{InFile}) or return 0; }
		my $data = 0;
		binmode($handle);
		while(<$handle>)
		{
			unless(ord() < 5)
			{
				my $l = $_;
				$self->{Buffer} .= $l;
				if(!$data)
				{
					if($self->LineCheck(\$l)){$data = 1;}
					else{ $self->MakeNaked(\$l); push(@{$self->{RawHeader}}, \$l); }
				}
				else{ last; }
			}
			else
			{
				my $ord = ord();
				if($ord == 2)
				{
					#reinit not necessary
					#$self->Init();
				}
				elsif($ord == 3 or $ord == 4) #end
				{
					if($ord == 4){ $handle eq STDIN or close($handle); }
					return $ord;
				}
			}
		}
		$handle eq STDIN or close($handle);
		return 1;
	}
	else{ return 0; }
}

sub Read
{
	my $self = shift;
	$self->Init(); #do we really want full init?
	$self->{InFile} = shift if defined $_[0]; 
	if(defined $self->{InFile})
	{
		my $handle;
		if($self->{InFile} eq ''){ $handle = STDIN; }
		else{ open($handle, $self->{InFile}) or return 0; }
		my $data = 0;
		binmode($handle);
		while(<$handle>)
		{
			unless(ord() < 5)
			{
				my $l = $_;
				if(!$data)
				{
					if($self->LineCheck(\$l)){$data = 1;}
					else{ $self->MakeNaked(\$l); push(@{$self->{RawHeader}}, \$l); }
				}
				if($data)
				{
					my $da = $self->LineData(\$l);
					push(@{$self->{Data}}, $da) if defined $da;
				}
			}
			else
			{
				my $ord = ord();
				if($ord == 2)
				{
					#reinit not necessary
					#$self->Init();
				}
				elsif($ord == 3 or $ord == 4) #end
				{
					if($ord == 4){ $handle eq STDIN or close($handle); }
					return $ord;
				}
			}
		}
		$handle eq STDIN or close($handle);
		return 1;
	}
	else{ return 0; }
}

#ComputeIndex(\arrayofcols)
sub ComputeIndex 
{
	my $self = shift;
	my $ar = shift;
	my $zahl = $self->{config}->{numregex};
	$ar = [0] unless defined $ar;
	foreach my $c (@{$ar}){ $self->{DataIndex}->[$c] = {};}
	foreach my $d (@{$self->{Data}})
	{
		foreach my $c (@{$ar})
		{
			my $key = $d->[$c] =~ /$zahl/ ? sprintf($self->{config}->{indexformat}, $d->[$c]) : $d->[$c];
			$self->{DataIndex}->[$c]->{$key} = $d unless defined $self->{DataIndex}->[$c]->{$key};
		}
	}
}

#(col)
sub ComputeSortedIndex
{
	my $self = shift;
	my $xl = shift;
	$xl = [0] unless defined $xl;
	#compute any missing indices
	my @l = ();
	foreach my $x (@{$xl}){	push(@l, $x) unless defined $self->{DataIndex}->[$x]; }
	$self->ComputeIndex(\@l);
	#now sort all concerned indices
	foreach my $x (@{$xl})
	{
		$self->{SortedData}->[$x] = [];
		foreach my $k (sort {$a <=> $b} keys %{$self->{DataIndex}->[$x]})
		{
			push(@{$self->{SortedData}->[$x]}, $self->{DataIndex}->[$x]->{$k});
		}
	}
}


sub WriteToHandle
{
	my $self = shift;
	my $handle = shift;
	my $selection = shift;
	#header
	$self->WriteHeader($handle, $selection);
	#data
	$self->WriteData($handle, $selection);
}

sub WriteData
{
	my $self = shift;
	my $handle = shift;
	my $selection = shift;
	for(my $i = 0; $i <= $#{$self->{Data}}; ++$i)
	{
		print $handle ${$self->DataLine($self->{Data}->[$i],$selection)};
	}
}

#Write(file, columns)
#maybe adding limits in future...
sub Write
{
	my $self = shift;
	$self->{OutFile} = $_[0] if defined $_[0];
	if(defined $self->{OutFile})
	{
		my $handle;
		if($self->{OutFile} eq ''){ $handle = STDOUT; }
		else{ open($handle, ">$self->{OutFile}") or return 0; }
		binmode($handle);
		$self->WriteToHandle($handle);
		$handle eq \*STDOUT or close($handle);
	}
	else{ return 0; }
}

sub WriteHeader
{
	my $self = shift;
	my $handle = shift;
	my $selection = shift;
	foreach my $h (@{$self->{RawHeader}}){ print $handle ${$self->CommentLine($h)}; }
	if(defined $selection)
	{
		print $handle ${$self->TitleLine($selection)};			
	}
}

sub WriteNewHeader
{
	my $self = shift;
	my $handle = shift;
	my $selection = shift;
	print $handle ${$self->CommentLine(\$self->{title})};
	foreach my $c (@{$self->{comments}}){ print $handle ${$self->CommentLine(\$c)}; }
	print $handle ${$self->TitleLine($selection)};			
}

sub SetOf
{
	my $self = shift;
	my $value = shift;
	my $zahl = $self->{config}->{numregex};
	$value = sprintf($self->{config}->{indexformat}, $value) if $value =~ /$zahl/;
	my $x = shift;
	$x = 0 unless defined $x;
	defined $self->{DataIndex}->[$x] or $self->ComputeIndex([$x]);
	return $self->{DataIndex}->[$x]->{$value};	
}

sub IntSetOf
{
	my $self = shift;
	#the easy way
	my $set = $self->SetOf(@_);
	return $set if defined $set;

	#so, we got to do some Work;
	my $val = shift;
	my $x = shift;
	$x = 0 unless defined $x;
	#find suitable data value pairs for interpolation
	my $sets = $self->Neighbours($val, $x);
	return undef unless defined $sets and $#{$sets} == 1;
	#compute
	my @set = ();

	for(my $y = 0; $y <= $#{$sets->[0]}; ++$y)
	{
		if($y != $x)
		{
			push(@set, LinearValue($val, [$sets->[0][$x], $sets->[1][$x]], [$sets->[0][$y],$sets->[1][$y]]));
		}
		else{ push(@set, $val); }
	}
	
	return \@set;
}

sub Y
{
	my $self = shift;
	my $val = shift;
	my $x = shift;
	my $y = shift;
	$x = 0 unless defined $x;
	$y = 1 unless defined $y;
	my $set = $self->SetOf($val,$x);
	return defined $set ? $set->[$y] : undef;
}


sub IntY
{
	my $self = shift;
	#the easy way
	my $y = $self->Y(@_);
	return $y if defined $y;

	#so, we got to do some Work;
	my $val = shift;
	my $x = shift;
	$y = shift;
	$x = 0 unless defined $x;
	$y = 1 unless defined $y;
	#find suitable data value pairs for interpolation
	my $sets = $self->Neighbours($val, $x);
	return undef unless defined $sets and $#{$sets} == 1;
	#compute
	#print STDERR "Sets: ";
	#for(@{$sets}){ print STDERR "(@{$_}) "; }
	#print STDERR "\n";
	return LinearValue($val, [$sets->[0][$x], $sets->[1][$x]], [$sets->[0][$y],$sets->[1][$y]]);
}

#it should be configurable wheter extrapolation is acceptable

sub Neighbours
{
	my $self = shift;
	my $val = shift;
	my $x = shift;
	
	my $n = undef;
	my $e = undef;

	unless($self->{config}{orderedint})
	{
		defined $self->{SortedData}->[$x] or $self->ComputeSortedIndex([$x]);
		return undef unless $#{$self->{SortedData}->[$x]} > 0; #senseless when not at least two points there

		foreach my $v (@{$self->{SortedData}->[$x]})
		{
			$n = $e;
			$e = $v;
			if($v->[$x] > $val and defined $n){ last; }
		}
	}
	else
	{
		if(defined $self->{Data}[0] and $self->{Data}[0][$x] < $val)
		{
			foreach my $v (@{$self->{Data}})
			{
				$n = $e;
				$e = $v;
				if($v->[$x] > $val and defined $n){ last; }
			}
		}
		else
		{
			foreach my $v (@{$self->{Data}})
			{
				$n = $e;
				$e = $v;
				if($v->[$x] < $val and defined $n){ last; }
			}
		}
	}
	return undef unless defined $n; #catches both  empty array and one-element-array
	unless($self->{config}{extrapol})
	{
		#check if we have left and right neighbour
		return undef unless (($n->[$x] <= $val and $e->[$x] >= $val) or ($n->[$x] >= $val and $e->[$x] <= $val))
	}
	return [$n,$e];
}


sub Max
{
	my $self = shift;
	my $formula = shift;
	my $ff = FormulaFunction($formula);
	return undef unless defined $ff;
	my $max = undef;
	foreach my $d (@{$self->{Data}})
	{
		my $nm = &$ff([$d]);
		$max = $nm if !defined $max or $nm > $max;
	}
	return $max;
}

sub Min
{
	my $self = shift;
	my $formula = shift;
	$formula = '-('.$formula.')';
	return -1*$self->Max($formula);
}

sub Sort
{
	my $self = shift;
	my $cols = shift;
	my $down = shift;
	my $sortcode = 'my $r; ';
	my $i = -1;
	# tested example sortcode:
	# '{ my $c; ($c= $a->[0] <=> $b->[0]) ? $c : ($c= $a->[1] <=> $b->[1]) ? $c : 0 }'
	foreach my $c (@{$cols})
	{
		my $r = '$r';
		my $a = '$a';
		my $b = '$b';
		if(defined $down and $down->[++$i])
		{
			$a = '$b';
			$b = '$a';
		}
		$sortcode .= "($r= $a\->[$c] <=> $b\->[$c]) ? $r : ";
	}
	$sortcode .= '0';
	# Not sure... is it better to eval that function here or to to put the eval into the sort call?
	print STDERR "Sort code: $sortcode\n";
	my $sortfunc = eval 'sub {'.$sortcode.'}';
	@{$self->{Data}} = sort $sortfunc @{$self->{Data}};
}

sub Calc
{
	my $self = shift;
	my $formula = shift;
	my $ff = FormulaFunction($formula);
	return undef unless defined $ff;
	foreach my $d (@{$self->{Data}})
	{
		&$ff([$d]);
	}
	return 1;
}

sub Mean
{
	my ($self, $col, $xcol, $begin, $end) = @_;
	my $count = 0;
	my $sum = 0;
	defined $self->{SortedData}->[$xcol] or $self->ComputeSortedIndex([$xcol]);
	foreach my $a (@{$self->{SortedData}->[$xcol]})
	{
		if($a->[$xcol] >= $start)
		{
			if($a->[$xcol] <= $end){ $sum += $a->[$col]; ++$count; }
			else{ last; }
		}
	}
	return $count > 0 ? $sum/$count : undef;
}

sub GetSortedData
{
	my $self = shift;
	my $x = shift;
	defined $self->{SortedData}->[$x] or $self->ComputeSortedIndex([$x]);
	return $self->{SortedData}->[$x];
}

sub Columns
{
	my $self = shift;
	return @{$self->{Data}} ? $#{$self->{Data}[0]}+1 : 0;
}

1;

__END__

=head1 NAME

TextData::File - process a whole file with text data

=head1 SYNOPSIS

	use TextData::File;

	#read $filename on construction
	my $file = new TextData::File(\%config,$filename);

	#create a fresh object without contents...
	my $file2 = new TextData::File(\%config);
	#...and read the file afterwards
	$file2->Read($filename);	

	print "third value of fourth data set: ",$file->{Data}->[3][2],"\n";	

=head1 DESCRIPTION

This is a subclass of TextData::Lines, so all properties are still there and only some sugar is added. It abstracts a file and contains all data for instant access (and memory consumption;-).
There is the possibility to address the data sets indexed via an arbitrary colum and to do some linear interpolation between points in this index (for that, the TextData::Calc package is needed).
A word on this index: It connects one value of the index column with the first data set it occured in - other sets containing this value are not concerned!

=head1 MEMBERS

=head2 Methods

=over * 4

=item * Init

erases data and parsed config stuff from memory; leaving only InFile and OutFile untouched

=item * ReadHead($file)

Just read head part of file (titles), closing file afterwards or slurping the header including first data line into the buffer.

=item * Read($file)

Read the data from the filename provided on construction or from $file if defined (where the internal value for the associated filename is set to $file). If $file is the empty string, data is read from STDIN.

=item * Write($file,\@selection)

writes the header and all data or columns in @selection to $file (use WriteFile(undef, \@selection) to write to the internally remembered output file from a previous run).

=item * WriteToHandle($handle, \@selection)

Does a full write to given file handle without closing it.

=item * WriteHeader($handle, \@selection)

writes a header to file handle $handle, tries to provide the appropriate column titles as last line when @selection is defined. Apart from this possibly constructed last line the header here is just the raw header read from the input file - including the original column titles if they were there.

=item * WriteData($handle, \@selection)

writes only the data part to file handle

=item * WriteNewHeader($handle,\@selection)

constructs a new header according to the selected columns, preserving the title and comments from the old header but putting it in current comment style and omitting the obsolete column title line, and writes it to $handle.

=item * SetOf($value,$indexcolumn) -> \@dataset

returns the data set (line) corresponding to $value in the index of $indexcolumn (0 is default if not specified)

=item * IntSetOf($value,$indexcolumn) -> \@dataset

does the same as above but applies linear interpolation to construct a data set for $value if it does not occur in the original data.

=item * Y($xvalue,$x,$y) -> $yvalue

returns a specific "Y" value to the "X" value $xvalue with the optional parameters $x and $y telling which columns we mean with X and Y - they default to 0 and 1; so the name of this method actually makes sense with files that contain X-Y data.

=item * IntY($xvalue,$x,$y) -> $yvalue

...see IntSetOf; just the same for Y()

=item * Neighbours($val,$x) -> \($n1,$n2)

searches two neighbouring points in column $x suitable for linear interpolation for $val.

=item * ComputeIndex(\@columns)

computes indexes (hash) with elements of the named columns as keys and the according data set (row) as value. Only the first found occurrence of the key in the data makes it into the index! This sets $file->{DataIndex}. 

=item * ComputeSortedIndex(\@columns)

computes the sorted indices for @columns (and the normal indices before that if necessary)

=item * GetSortedData($column) -> \@sorted_data

returns ref to array of all data sets that made it into the index for $column in ascending order in respect to $column.

=item * Max($formula) -> $maxval

computes the maximum value of the given formula over all data sets

=item * Min($formula) -> $minval

computes the minimum value of the given formula over all data sets

=item * MaxVal($col) NOT IMPLEMENTED

This may be a faster specialization of Max("[$col]")...

=item * Sort(\@cols, \@down)

Sort the data in-place in a stable manner, according to column indices @cols, in order. Optional boolean array @down decides if we sort down or up (default) for each column.

=item * Calc($formula)

Perform given Calculation over all data sets.

=item * Mean($col, $xcol, $begin, $end) -> $mean_value

calculates the arithmetic mean value of column $col taken according the given range $begin to $end of column $xcol

=item * Columns()

Give number of columns in data.

=back

=head2 Data

Here are some keys for the hash reference representing a TextData::File.

=over * 4

=item * Data

2-dimensional array reference. Stores values as $file->{Data}->[$set][$col] (data set / line and column start at 0).

=item * DataIndex

An alternative access for the data sets; $file->{DataIndex}->[$indexcol]->{$value} is a reference to the data set where $value stands in column $indexcol (in fact only the first occurence).

=item * SortedData

@{$file->{SortedData}->[$x]} contains the data sets that made it into the index of column $x in ascending order concerning the values of column $x.

=item * Buffer

A raw string buffer holding file (stdin) containing what has been read by ReadHead and may be inaccessible otherwise.

=back

Additional keys in $self->{config}:

=over * 4 

=item * indexformat

=item * orderedint (0)

Normally, interpolation is made between the nearest neighbours of the desired value - determined out of the whole file with SortedIndex. If this option is true, however, the data sets are searched in the true order of the file, thus interpolation takes place between the first two points that appear to enclose the wanted one. This may be preferrable in some situations where the data is not monotonic in the variable used as index.

=item * extrapol (1)

allows automatic extrapolation

=back

=head1 LICENSE

This package is open source software under the terms of the Artistic License. Free use respecting the author;-)

=head1 AUTHOR

Thomas Orgis <thomas@orgis.org>

=head1 WEB

See	http://thomas.orgis.org/scripterei for updates and lots of other (un)related stuff.
